import { singleton } from '@/decorators/singleton.decorator'
import WalletConnectProvider from '@walletconnect/web3-provider'
import { AppModule } from '@/store/modules/app'
import { Contract } from 'web3-eth-contract'
import Web3 from 'web3'
import {
  AVAILABLE_CHAINS,
  CONTRACT_ADDRESS,
  IChainData
} from '@/configs/config_chain'
import {
  EventBus,
  NEED_CHANGE_CHAIN,
  NEED_LOGIN,
  NEED_NONCE,
  SHOW_CHAIN_MODAL
} from '@/utils/event-bus'
import { UserModule } from '@/store/modules/user'
import { isMobile } from '@/utils/resize'

import { AllChains } from '@/configs/allchain'
import { MessageBox } from 'element-ui'
import { ERC20ABI } from '@/configs/contracts'
import { TransactionReceipt } from 'web3-core'
import { hasMetamask, toHexChainId } from '@/utils/chain.util'

const EIP721_DOMAIN_DATA = [
  { name: 'name', type: 'string' },
  { name: 'version', type: 'string' },
  { name: 'chainId', type: 'uint256' },
  { name: 'verifyingContract', type: 'address' }
]

const CACHE_KEY = 'cebg_chain_cache_key'

@singleton
export class Blockchain {
  provider: any
  web3: Web3 | null = null
  currentChain = 321
  supportChain = AVAILABLE_CHAINS
  // 0: null, 1: metamask, 2: walletconnect
  walletType = 0
  dataCached = false
  instanceCacheMap: Map<string, any>
  public chainMap: Map<number, IChainData> = new Map()
  public rpc: any = {}

  constructor() {
    const allChainMap: Map<number, IChainData> = new Map()
    for (const d of AllChains) {
      const id = d.id
      this.rpc[id] = d.rpc
      allChainMap.set(id, d)
    }
    this.chainMap.set(this.supportChain, allChainMap.get(this.supportChain)!)

    // this.loadCachedProvider()
    this.instanceCacheMap = new Map()
    // AppModule.updateChainID(chainId)
    EventBus.$on(NEED_LOGIN, this.connect.bind(this))
  }

  get isWalletConnect() {
    return !!this.walletType && !!this.currentChain
  }

  get mallAddress() {
    return CONTRACT_ADDRESS[this.currentChain]?.mall
  }

  public get hexChainId() {
    return toHexChainId(this.currentChain)
  }

  public async chainSelected(id: number) {
    if (!this.chainMap.has(id)) {
      return
    }
    this.currentChain = id
    if (this.provider) {
      await this.switchEthereumChain()
    } else {
      await this.connectWallet()
    }
  }

  public async connectWallet() {
    console.log('begin connect to wallet: ', this.walletType, this.currentChain)
    if (this.walletType === 1) {
      this.provider = await this.connectMetaMask()
    } else if (this.walletType === 2) {
      this.provider = await this.connectWalletConnect()
    }
    if (!this.provider) {
      return
    }
    this.web3 = new Web3(this.provider)
    let chainId = await this.web3.eth.getChainId()

    await this.checkChain(chainId)

    this.subscribeToEvents()
    const accounts = await this.web3.eth.getAccounts()
    if (accounts && accounts.length > 0) {
      UserModule.updateAccount(accounts[0])
    }
    if (!this.currentChain) this.currentChain = chainId
    this.saveProvider()
    UserModule.updateChainID(chainId)
    AppModule.updateWalletStatus(true)
    console.log('current login chain: ', chainId)
    console.log('accountsLogin: ', accounts, UserModule.accountId)
    EventBus.$emit(NEED_NONCE)
    return { account: accounts[0], chainId }
  }

  /**
   * check whether special chainId is supported by config
   * show chain picker when chainId is not supported and current wallet is meatmask
   * show message box when chainId is not supported and current wallet is wallet connect
   * @param {number} chainId
   * @return {Promise<void>}
   * @private
   */
  private async checkChain(chainId: number) {
    console.log('chechain', this.chainMap)
    if (!this.chainMap.has(chainId)) {
      try {
        await this.selectChain()
      } catch (err) {
        await this.disconnect()
      }
    }
  }

  public async connect(walletType: number) {
    console.log('blockchain connect: ' + walletType)
    this.walletType = walletType || this.walletType
    await this.connectWallet()
  }

  /**
   * show wallet picker
   * @return {Promise<number>}
   * @private
   */
  private selectWallet(): Promise<number> {
    return new Promise((resolve, reject) => {
      EventBus.$emit(SHOW_CHAIN_MODAL, {
        confirm: (id: number) => {
          console.log('select wallet: ', id)
          resolve && resolve(id)
        },
        cancel: (reason: any) => {
          console.log('cancel select wallet: ', reason)
          reject && reject(reason)
        }
      })
    })
  }

  /**
   * show chain picker
   * @return {Promise<number>}
   * @private
   */
  private selectChain(): Promise<number> {
    let tarChain = this.chainMap.get(this.supportChain)
    return new Promise((resolve, reject) => {
      MessageBox.confirm(`确定切换到 ${tarChain?.name}?`, '提示', {
        confirmButtonText: 'OK',
        cancelButtonText: 'Cancel',
        type: 'info'
      })
        .then(async () => {
          console.log('select chain:', tarChain?.id)
          this.currentChain = this.supportChain
          if (this.provider) {
            await this.switchEthereumChain()
          }
          resolve && resolve(this.supportChain)
        })
        .catch(() => {
          reject && reject('')
        })
    })
  }

  public async connectMetaMask() {
    let provider = null
    if (typeof window.ethereum !== 'undefined') {
      provider = window.ethereum
      try {
        await provider.request({ method: 'eth_requestAccounts' })
      } catch (error) {
        // @ts-ignore
        if (error.code === -32002) {
          // const hexChainId = toHexChainId(this.currentChain)
          // await provider.request({
          //   method: 'wallet_switchEthereumChain',
          //   params: [{ chainId: hexChainId }]
          // })
          throw new Error('MeatMask not login, Open MeatMask and login first')
        } else {
          throw new Error('User Rejected')
        }
      }
    } else if (window.web3) {
      provider = window.web3.currentProvider
    } else if (window.celo) {
      provider = window.celo
    } else {
      throw new Error('No Web3 Provider found')
    }
    return provider
  }

  public async connectWalletConnect() {
    const provider = new WalletConnectProvider({
      infuraId: process.env.VUE_APP_WALLET_INFURAID,
      rpc: this.rpc
    })
    try {
      await provider.enable()
    } catch (err) {
      console.log('connect to wallet connect error: ', err)
      await Promise.reject(err)
    }

    return provider
  }

  public async getContractInstance(address: string, abi: any = ERC20ABI) {
    if (!this.instanceCacheMap.has(address)) {
      const instance = new Contract(abi, address, {
        from: UserModule.accountId
      })
      this.instanceCacheMap.set(address, instance)
    }
    return this.instanceCacheMap.get(address)
  }

  public clearCachedProvider() {
    console.log('clear cached provider')
    localStorage.removeItem(CACHE_KEY)
  }

  public loadCachedProvider() {
    const dataStr = localStorage.getItem(CACHE_KEY)
    if (dataStr) {
      try {
        const data = JSON.parse(dataStr)
        this.walletType = data.walletType
        if (this.chainMap.has(data.chainId)) {
          this.currentChain = data.chainId
        }
      } catch (err) {
        console.log('err parse cached json')
        this.clearCachedProvider()
      }
    }
  }

  public saveProvider() {
    const data = {
      walletType: this.walletType,
      chainId: this.currentChain
    }
    localStorage.setItem(CACHE_KEY, JSON.stringify(data))
  }

  public async disconnect() {
    try {
      await UserModule.LogOut()
      await this.provider?.disconnect()
    } catch (err) {}
    this.currentChain = 0
    this.walletType = 0
    this.instanceCacheMap.clear()
    this.clearCachedProvider()
    AppModule.updateStep(0)
    UserModule.updateChainID(0)
    AppModule.updateNonce('')
    UserModule.updateAccount('')
    AppModule.updateWalletStatus(false)
  }

  public subscribeToEvents = () => {
    // Subscribe to accounts change
    this.provider.on('accountsChanged', async (accounts: string[]) => {
      console.log('accountsChanged: ', accounts)
      if (accounts && accounts.length > 0) {
        if (UserModule.accountId !== accounts[0]) {
          await UserModule.LogOut()
          location.reload()
        }
      }
    })

    // Subscribe to chainId change
    this.provider.on('chainChanged', async (chainId: string) => {
      const chainIdNum = parseInt(chainId)
      console.log('chainChanged', chainId, chainIdNum)
      await this.checkChain(chainIdNum)
      this.currentChain = chainIdNum
      this.saveProvider()
      UserModule.updateChainID(this.currentChain)
    })

    // Subscribe to session disconnection
    this.provider.on('disconnect', (err: any) => {
      console.log('disconnect', err)
    })
  }

  /**
   * change chain of metamask
   * @param {number} chainId
   * @param {() => void} cb
   * @return {Promise<void>}
   */
  async switchEthereumChain(chainId?: number, cb?: () => void) {
    chainId = chainId || this.currentChain
    const hexChainId = toHexChainId(chainId)
    const onChainChange = (chainId: string) => {
      console.log('switchEthereumChain: ', chainId)
      this.provider.removeListener('chainChanged', onChainChange)
      cb && cb()
    }
    this.provider.on('chainChanged', onChainChange)
    try {
      await this.provider.request({
        method: 'wallet_switchEthereumChain',
        params: [{ chainId: hexChainId }]
      })
      console.log('switch chain success')
    } catch (e) {
      console.log('switch chain error: ', e)
      if (e.code === 4902 || e.message.indexOf('Unrecognized chain ID') >= 0) {
        try {
          const data = this.chainMap.get(chainId)!
          await this.provider.request({
            method: 'wallet_addEthereumChain',
            params: [
              {
                chainId: hexChainId,
                chainName: data.name,
                nativeCurrency: {
                  name: data.symbol,
                  symbol: data.symbol,
                  decimals: data.decimals || 18
                },
                blockExplorerUrls: [data.explorerurl],
                rpcUrls: [data.rpc]
              }
            ]
          })
          console.log('add chain success')
        } catch (addError) {
          console.error('add chain error: ', addError)
          this.provider.removeListener('chainChanged', onChainChange)
        }
      }
      // console.error(e)
    }
  }

  public async getTransactionReceipt(txHash: string) {
    return this.web3!.eth.getTransactionReceipt(txHash)
  }

  public async getTxConfirms(txhash: string) {
    const receipt: TransactionReceipt = await this.getTransactionReceipt(txhash)
    const latest = await this.web3!.eth.getBlockNumber()
    return latest - receipt.blockNumber + 1
  }

  public async signPresale({
    type,
    paymentTokenAddress,
    price,
    buyerAddress,
    netId
  }: {
    type: number | string
    paymentTokenAddress: string
    price: any
    buyerAddress: string
    netId?: any
  }) {
    const nonce = (Math.random() * 100000) | 0

    const signMsg = {
      item: type,
      token: paymentTokenAddress,
      price: price,
      salt: nonce
    }
    netId = netId || (await this.web3!.eth.getChainId())
    const signObj = {
      types: {
        EIP712Domain: EIP721_DOMAIN_DATA,
        set: [
          { name: 'item', type: 'uint256' },
          { name: 'token', type: 'address' },
          { name: 'price', type: 'uint256' },
          { name: 'salt', type: 'uint256' }
        ]
      },
      primaryType: 'set',
      domain: {
        name: 'BEBoxMall',
        version: '1',
        chainId: netId,
        verifyingContract: this.mallAddress
      },
      message: signMsg
    }
    const signature = await this.signData(signObj, buyerAddress)
    return { nonce, signature }
  }

  public async increaseAllowance(address: string, price: string) {
    const coinInstance: any = await this.getContractInstance(address)
    const res = await coinInstance.methods
      .increaseAllowance(this.mallAddress, price)
      .send({ gas: 1000000 })
    console.log('increaseAllowance: ', res)
  }

  /**
   * @param {string} address
   * @param {string | null} account
   * @return {Promise<any>}
   */
  public async getBalance(address: string, account: string | null) {
    account = account || UserModule.accountId
    const coinInstance: any = await this.getContractInstance(address)
    return await coinInstance.methods.balanceOf(account).call()
  }

  public async transferToAccount(
    account: string,
    amount: number,
    address: string
  ) {
    const amountBN = this.web3!.utils.toBN(this.web3!.utils.toWei(amount + ''))
    const coinInstance: any = await this.getContractInstance(address)
    return coinInstance.methods
      .transfer(account, amountBN)
      .send({ gas: 1000000 })
  }

  public async signData(signObj: any, signer: string) {
    const msgParams = JSON.stringify(signObj)
    const from = signer
    console.log('clicked, sending personal sign req', 'from', from, msgParams)
    const params = [from, msgParams]
    const result: any = await this.sendCmd({
      method: 'eth_signTypedData_v4',
      params,
      from
    })
    console.log(result)
    return result.result
  }

  public async sendCmd({ method, params, from }: any) {
    return new Promise((resolve, reject) => {
      // eslint-disable-next-line @typescript-eslint/ban-ts-comment
      // @ts-ignore
      this.web3.currentProvider.sendAsync(
        {
          method,
          params,
          from
        },
        async function(err: any, result: any) {
          if (err) {
            reject && reject(err)
            return
          }
          resolve && resolve(result)
        }
      )
    })
  }
}
