import {
  Action,
  getModule,
  Module,
  Mutation,
  VuexModule
} from 'vuex-module-decorators'
import { getAdminInfo, login, loginWallet, logout } from '@/api/admins'
import { getToken, removeToken, setToken } from '@/utils/cookies'
import store from '@/store'
import router, { resetRouter } from '@/router'
import { PermissionModule } from '@/store/modules/permission'
import { TagsViewModule } from '@/store/modules/tags-view'
import { RouteConfig } from 'vue-router'
import Layout from '@/layout/index.vue'

export interface IUserState {
  id: string
  token: string
  name: string
  avatar: string
  accountId?: string
  introduction: string
  roles: string[]
  email: string
  permissions: string[][]
  department?: string
  deptname?: string
  level: number
  sex?: string
  chainId?: number
}

const EIP721_DOMAIN_DATA = [
  { name: 'name', type: 'string' },
  { name: 'version', type: 'string' }
]

@Module({ dynamic: true, store, name: 'user' })
class User extends VuexModule implements IUserState {
  public token = getToken() || ''
  public id = ''
  public name = ''
  public avatar = ''
  public introduction = ''
  public roles: string[] = []
  public permissions: string[][] = []
  public email = ''
  public department = ''
  public deptname = ''
  public level = 999
  public sex = '0'
  public accountId = ''
  public chainId = 0

  @Action
  public async Login({
    username,
    password
  }: {
    username: string
    password: string
  }) {
    username = username.trim()
    const { data } = await login({ username, password })
    setToken(data.token)
    this.SET_TOKEN(data.token)
  }

  @Action({ rawError: true })
  public async WalletLogin({
    bcInstance,
    account,
    chainId,
    nonce,
    tips
  }: {
    bcInstance: any
    account: string
    chainId: string | number
    nonce: string
    tips?: string
  }) {
    nonce += ''
    tips = tips || 'This signature is only used for verify your account'
    const signMsg = {
      tips,
      nonce
    }

    const signObj = {
      types: {
        EIP712Domain: EIP721_DOMAIN_DATA,
        set: [
          { name: 'tips', type: 'string' },
          { name: 'nonce', type: 'string' }
        ]
      },
      primaryType: 'set',
      domain: {
        name: 'Auth',
        version: '1'
      },
      message: signMsg
    }
    const signature = await bcInstance.signData(signObj, account)
    const authData = {
      account,
      nonce,
      signature,
      tips,
      net_id: chainId
    }
    console.log('login data: ', authData)
    const res: any = await loginWallet(authData)

    console.log('login res: ', res)
    if (!res.code && res.data && res.data.token) {
      setToken(res.data.token)
      this.SET_TOKEN(res.data.token)
    }
  }

  @Action
  public ResetToken() {
    removeToken()
    this.SET_TOKEN('')
    this.SET_ROLES([])
    this.SET_PERMISSIONS([])
  }

  @Action
  public async GetUserInfo() {
    if (this.token === '') {
      throw Error('GetUserInfo: token is undefined!')
    }
    const { data } = await getAdminInfo({
      /* Your params here */
    })
    if (!data) {
      throw Error('Verification failed, please Login again.')
    }
    const {
      id,
      roles,
      showname,
      avatar,
      introduction,
      permissions,
      department,
      level,
      deptname,
      sex,
      flowRouter
    } = data
    // roles must be a non-empty array
    if (!roles || roles.length <= 0) {
      throw Error('GetUserInfo: roles must be a non-null array!')
    }
    this.SET_ID(id)
    this.SET_ROLES(roles)
    this.SET_NAME(showname)
    this.SET_AVATAR(avatar)
    this.SET_INTRODUCTION(introduction)
    this.SET_PERMISSIONS(permissions)
    this.SET_DEPARTMENT(department)
    this.SET_LEVEL(level)
    this.SET_DEPTNAME(deptname)
    this.SET_SEX(sex)
    return flowRouter
  }

  @Action
  public async UpdateInfo(data: any) {
    const { showname, sex, avatar } = data
    this.SET_NAME(showname)
    this.SET_SEX(sex)
    this.SET_AVATAR(avatar)
  }

  @Action
  public async ChangeRoles(role: string) {
    // Dynamically modify permissions
    const token = role + '-token'
    this.SET_TOKEN(token)
    setToken(token)
    const flowRouter = await this.GetUserInfo()
    resetRouter()
    // Generate dynamic accessible routes based on roles
    // TODO:: 根据实际接口返回, 删除测试数据
    const extRouter: RouteConfig = {
      path: '/system1',
      component: Layout,
      meta: {
        title: 'flow_begin',
        icon: 'component',
        alwaysShow: true
      },
      children: []
    }
    extRouter.children?.push({
      path: 'flow_begin',
      component: () => import('@/views/error-page/401.vue'),
      name: 'FlowBegin',
      meta: {
        title: 'flow_begin',
        permissions: ['flow_exec:read'],
        icon: 'game'
      }
    })
    flowRouter.forEach((o: any) => {
      extRouter.children?.push(o)
    })
    PermissionModule.GenerateRoutes({ extRouter, roles: this.roles })
    // Add generated routes
    router.addRoutes(PermissionModule.dynamicRoutes)
    // Reset visited views and cached views
    TagsViewModule.delAllViews()
  }

  @Action
  public async LogOut() {
    if (this.token === '') {
      throw Error('LogOut: token is undefined!')
    }
    await logout()
    removeToken()
    this.SET_TOKEN('')
    this.SET_ROLES([])
    this.SET_PERMISSIONS([])
    this.SET_DEPARTMENT('')
  }

  @Mutation
  private SET_TOKEN(token: string) {
    this.token = token
  }

  @Action
  public async updatePageToken(token: string) {
    this.SET_TOKEN(token)
  }

  @Mutation
  private SET_NAME(name: string) {
    this.name = name
  }

  @Mutation
  private SET_AVATAR(avatar: string) {
    this.avatar = avatar
  }

  @Mutation
  private SET_INTRODUCTION(introduction: string) {
    this.introduction = introduction
  }

  @Mutation
  private SET_ROLES(roles: string[]) {
    this.roles = roles
  }
  @Mutation
  private SET_ID(id: string) {
    this.id = id
  }

  @Mutation
  private SET_DEPARTMENT(department: string) {
    this.department = department
  }

  @Mutation
  private SET_DEPTNAME(deptname: string) {
    this.deptname = deptname
  }

  @Mutation
  private SET_SEX(sex: string) {
    this.sex = sex
  }

  @Mutation
  private SET_LEVEL(level: number) {
    this.level = level
  }

  @Mutation
  private SET_PERMISSIONS(permissions: string[]) {
    const results: string[][] = []
    for (const permission of permissions) {
      if (permission === '*') {
        results.push(['*', '*'])
      } else {
        results.push(permission.split(':'))
      }
    }
    this.permissions = results
  }

  @Action
  public updateAccount(val: string) {
    this.UPDATE_ACCOUNT(val)
  }

  @Action
  public updateChainID(val: number) {
    this.UPDATE_CHAINID(val)
  }

  @Mutation
  private UPDATE_ACCOUNT(val: string) {
    this.accountId = val
  }

  @Mutation
  private UPDATE_CHAINID(val: number) {
    this.chainId = val
  }
}

export const UserModule = getModule(User)
