interface Date {
  /**
   * 增加n天
   * @param {number} days
   * @return {Date}
   */
  addDays(days: number): Date
}
// eslint-disable-next-line no-extend-native
Object.defineProperties(Date.prototype, {
  addDays: {
    value: function(days: number) {
      return new Date(this.getTime() + days * 3600 * 24 * 1000)
    },
    writable: true
  }
})
interface CanvasRenderingContext2D {
  roundRect(x: number, y: number, w: number, h: number, r: number):CanvasRenderingContext2D
}
Object.defineProperties(CanvasRenderingContext2D.prototype, {
  roundRect: {
    value: function(x: number, y: number, w: number, h: number, r: number) {
      const minSize = Math.min(w, h)
      if (r > minSize / 2) r = minSize / 2
      // 开始绘制
      this.beginPath()
      this.moveTo(x + r, y)
      this.arcTo(x + w, y, x + w, y + h, r)
      this.arcTo(x + w, y + h, x, y + h, r)
      this.arcTo(x, y + h, x, y, r)
      this.arcTo(x, y, x + w, y, r)
      this.strokeStyle = '#FFF'
      this.stroke()
      this.closePath()
      return this
    },
    writable: true
  }
})

interface Array<T> {
  /**
   * 如果数组中没有要放入的对象，则将对象放入数组
   *
   * @param {T} t 要放入的对象
   * @returns {number} 放入的对象，在数组中的索引
   *
   * @member Array
   */
  pushOnce(t: T): number

  /**
   *
   * 删除某个数据
   * @param {T} t
   * @returns {boolean}   true 有这个数据并且删除成功
   *                      false 没有这个数据
   */
  zremove(t: T): boolean

  /**
   * 将数组克隆到to
   * to的数组长度会和当前数组一致
   *
   * @template T
   * @param {Array<T>} to
   */
  cloneTo<T>(to: Array<T>): void

  /**
   * 将数组附加到to中
   *
   * @template T
   * @param {Array<T>} to
   *
   * @member ArrayConstructor
   */
  appendTo<T>(to: Array<T>): void

  /**
   * 移除数组index位置的元素, 比slice效率高
   * @param index
   */
  spliceOne(index: number): boolean

  /**
   * 随机排序
   */
  randomSort(): void

  /**
   * 检查数组中是否含有另外一个object
   * @param obj 与数组同类型的obj | 同类型的数组 | 指定child字段的值 | 指定child字段的数组
   * @param child 比较字段
   */
  contains<T>(obj: T | T[] | {} | {}[], child?: string): boolean

  /**
   * 将数组随机插入当前数组中
   * @param arr
   */
  randomInsert<T>(arr: Array<T>): void

  /**
   * 随机获取n个元素
   * @param count
   */
  randomGet<T>(count?: number): T[]

  /**
   * 随机获取1个元素
   */
  randomOne<T>(): T

  /**
   * 随机移除n个元素
   * @param count
   */
  randomRemove<T>(count?: number): T[]

  /**
   * 数组移动n位
   * @param n n > 0 右移, n<0 左移
   */
  moveElement<T>(n: number): T[]

  /**
   * 两个数组并集
   * @param arr
   */
  union<T>(arr: T[]): T[]

  /**
   * 两个数组交集
   * @param arr
   */
  intersect<T>(arr: T[]): T[]

  /**
   * 相对于arr的差集
   * @param arr
   */
  difference<T>(arr: T[]): T[]

}

// eslint-disable-next-line no-extend-native
Object.defineProperties(Array.prototype, {
  cloneTo: {
    value: function <T>(this: T[], b: any[]) {
      b.length = this.length
      const len = this.length
      b.length = len
      for (let i = 0; i < len; i++) {
        b[i] = this[i]
      }
    },
    writable: true
  },
  appendTo: {
    value: function <T>(this: T[], b: any[]) {
      const len = this.length
      for (let i = 0; i < len; i++) {
        b.push(this[i])
      }
    },
    writable: true
  },
  pushOnce: {
    value: function <T>(this: T[], t: T) {
      let idx = this.indexOf(t)
      if (!~idx) {
        idx = this.length
        this.push(t)
      }
      return idx
    },
    writable: true
  },
  zremove: {
    value: function <T>(this: T[], t: T) {
      const idx = this.indexOf(t)
      if (~idx) {
        this.splice(idx, 1)
        return true
      }
      return false
    },
    writable: true
  },
  spliceOne: {
    value: function(index: number): boolean {
      if (index === -1 || index >= this.length) {
        return false
      }
      const len = this.length - 1
      for (let i = index; i < len; i++) {
        this[i] = this[i + 1]
      }
      this.length = len
      return true
    },
    writable: true
  },

  randomSort: {
    value: function <T>() {
      for (let j, x, i = this.length; i; j = (Math.random() * i) | 0, x = this[--i], this[i] = this[j], this[j] = x) {}
    },
    writable: true
  },
  randomInsert: {
    value: function <T>(arr: Array<T>) {
      const length = this.length
      arr.forEach(value => {
        this.splice(Math.random() * length, 0, value)
      })
    },
    writable: true
  },
  randomGet: {
    value: function <T>(count = 1): T[] {
      const shuffled: T[] = this.slice(0)
      let i = this.length
      const min = i - count
      let temp
      let index
      if (min < 0) {
        return shuffled
      }
      while (i-- > min) {
        index = Math.floor((i + 1) * Math.random())
        temp = shuffled[index]
        shuffled[index] = shuffled[i]
        shuffled[i] = temp
      }
      return shuffled.slice(min)
    },
    writable: true
  },
  randomOne: {
    value: function <T>(): T|null {
      const results = this.randomGet(1)
      if (results.length > 0) {
        return results[0]
      } else {
        return null
      }
    },
    writable: true
  },
  randomRemove: {
    value: function <T>(count = 1): T[] {
      const result = []
      while (count-- > 0 && this.length > 0) {
        const index = (Math.random() * this.length) | 0
        result.push(...this.splice(index, 1))
      }
      return result
    },
    writable: true
  },

  moveElement: {
    value: function <T>(n: number): T[] {
      if (Math.abs(n) > this.length) n = n % this.length
      return this.slice(-n).concat(this.slice(0, -n))
    },
    writable: true
  },

  union: {
    value: function <T>(this: T[], b: any[]): T[] {
      const a = this.concat(b)
      return [...new Set(a)]
    },
    writable: true
  },

  intersect: {
    value: function <T>(this: T[], b: any[]): T[] {
      const set0 = new Set(b)
      const set1 = new Set(this.filter(x => set0.has(x)))
      return [...set1]
    },
    writable: true
  },

  difference: {
    value: function <T>(this: T[], b: any[]): T[] {
      const set0 = new Set(b)
      const set1 = new Set(this.filter(x => !set0.has(x)))
      return [...set1]
    },
    writable: true
  }
})

interface Map<K, V> {
  /**
   * 只针对V为number的Map, 有值的话, 加上V, 没值则直接set
   * V为其他类型时, 直接set
   * @param key
   * @param value
   */
  inc(key: K, value: V): number
}

// eslint-disable-next-line no-extend-native
Object.defineProperties(Map.prototype, {
  inc: {
    value: function <K, V>(key: K, value: V) {
      if (typeof value === 'number') {
        this.set(key, (this.get(key) || 0) + value)
      } else {
        this.set(key, value)
      }
      return this.get(key)
    }
  }
})
